#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
udocker unit tests: UdockerCLI
"""

from unittest import TestCase, main
from unittest.mock import Mock, patch
from udocker.config import Config
from udocker.cmdparser import CmdParser
from udocker.cli import UdockerCLI

BUILTIN = "builtins"
BOPEN = BUILTIN + '.open'


class UdockerCLITestCase(TestCase):
    """Test UdockerTestCase() command line interface."""

    def setUp(self):
        Config().getconf()
        Config().conf['hostauth_list'] = ("/etc/passwd", "/etc/group")
        Config().conf['cmd'] = "/bin/bash"
        Config().conf['cpu_affinity_exec_tools'] = \
            (["numactl", "-C", "%s", "--", ],
             ["taskset", "-c", "%s", ])
        Config().conf['valid_host_env'] = "HOME"
        Config().conf['username'] = "user"
        Config().conf['userhome'] = "/"
        Config().conf['oskernel'] = "4.8.13"
        Config().conf['location'] = ""
        Config().conf['keystore'] = "KEYSTORE"

        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

    def tearDown(self):
        self.lrepo.stop()

    @patch('udocker.cli.LocalFileAPI')
    @patch('udocker.cli.KeyStore')
    @patch('udocker.cli.DockerIoAPI')
    def test_01_init(self, mock_dioapi, mock_ks, mock_lfapi):
        """Test01 UdockerCLI() constructor."""
        # Test Config().conf['keystore'] starts with /
        Config().conf['keystore'] = "/xxx"
        UdockerCLI(self.local)
        self.assertTrue(mock_dioapi.called)
        self.assertTrue(mock_lfapi.called)
        self.assertTrue(mock_ks.called_with(Config().conf['keystore']))

        # Test Config().conf['keystore'] does not starts with /
        Config().conf['keystore'] = "xx"
        UdockerCLI(self.local)
        self.assertTrue(mock_ks.called_with(Config().conf['keystore']))

    @patch('udocker.cli.FileUtil.isdir')
    def test_02__cdrepo(self, mock_isdir):
        """Test02 UdockerCLI()._cdrepo()."""
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc._cdrepo(cmdp)
        self.assertFalse(status)
        self.assertFalse(mock_isdir.called)

        argv = ["udocker"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_isdir.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc._cdrepo(cmdp)
        self.assertFalse(status)
        self.assertTrue(mock_isdir.called)

        argv = ["udocker"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_isdir.return_value = True
        self.local.setup.return_value = None
        udoc = UdockerCLI(self.local)
        status = udoc._cdrepo(cmdp)
        self.assertTrue(status)
        self.assertTrue(self.local.setup.called)

    @patch('udocker.cli.DockerIoAPI.is_repo_name')
    @patch('udocker.cli.Msg')
    def test_03__check_imagespec(self, mock_msg, mock_reponame):
        """Test03 UdockerCLI()._check_imagespec()."""
        mock_msg.level = 0
        mock_reponame.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc._check_imagespec("")
        self.assertEqual(status, (None, None))

        mock_reponame.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc._check_imagespec("AAA")
        self.assertEqual(status, ("AAA", "latest"))

        mock_reponame.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc._check_imagespec("AAA:45")
        self.assertEqual(status, ("AAA", "45"))

    @patch('udocker.cli.DockerIoAPI.is_repo_name')
    @patch('udocker.cli.Msg')
    def test_04__check_imagerepo(self, mock_msg, mock_reponame):
        """Test04 UdockerCLI()._check_imagerepo()."""
        mock_msg.level = 0
        mock_reponame.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc._check_imagerepo("")
        self.assertEqual(status, None)

        mock_reponame.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc._check_imagerepo("AAA")
        self.assertEqual(status, "AAA")

    @patch('udocker.cli.DockerIoAPI.set_index')
    @patch('udocker.cli.DockerIoAPI.set_registry')
    @patch('udocker.cli.DockerIoAPI.set_proxy')
    @patch('udocker.cli.Msg')
    def test_05__set_repository(self, mock_msg, mock_proxy,
                                mock_reg, mock_idx):
        """Test05 UdockerCLI()._set_repository()."""
        mock_msg.level = 0
        regist = "registry.io"
        idxurl = "dockerhub.io"
        imgrepo = "dockerhub.io/myimg:1.2"
        mock_proxy.return_value = None
        mock_reg.side_effect = [None, None, None, None]
        mock_idx.side_effect = [None, None, None, None]
        udoc = UdockerCLI(self.local)
        status = udoc._set_repository(regist, idxurl, imgrepo, True)
        self.assertTrue(status)
        self.assertTrue(mock_proxy.called)
        self.assertTrue(mock_reg.called)
        self.assertTrue(mock_idx.called)

        regist = ""
        idxurl = ""
        imgrepo = "https://dockerhub.io/myimg:1.2"
        mock_proxy.return_value = None
        mock_reg.side_effect = [None, None, None, None]
        mock_idx.side_effect = [None, None, None, None]
        udoc = UdockerCLI(self.local)
        status = udoc._set_repository(regist, idxurl, imgrepo, False)
        self.assertTrue(status)

    def test_06__split_imagespec(self):
        """Test06 UdockerCLI()._split_imagespec()."""
        imgrepo = ""
        res = ("", "", "", "")
        udoc = UdockerCLI(self.local)
        status = udoc._split_imagespec(imgrepo)
        self.assertEqual(status, res)

        imgrepo = "dockerhub.io/myimg:1.2"
        res = ("", "dockerhub.io", "myimg", "1.2")
        udoc = UdockerCLI(self.local)
        status = udoc._split_imagespec(imgrepo)
        self.assertEqual(status, res)

        imgrepo = "https://dockerhub.io/myimg:1.2"
        res = ("https:", "dockerhub.io", "myimg", "1.2")
        udoc = UdockerCLI(self.local)
        status = udoc._split_imagespec(imgrepo)
        self.assertEqual(status, res)

    @patch('udocker.cli.os.path.exists')
    @patch('udocker.cli.Msg')
    def test_07_do_mkrepo(self, mock_msg, mock_exists):
        """Test07 UdockerCLI().do_mkrepo()."""
        mock_msg.level = 0

        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_exists.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_mkrepo(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(mock_exists.called)

        argv = ["udocker", "mkrepo"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_exists.return_value = False
        self.local.setup.return_value = None
        self.local.create_repo.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_mkrepo(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_exists.called)
        self.assertTrue(self.local.setup.called)
        self.assertTrue(self.local.create_repo.called)

        argv = ["udocker", "mkrepo"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_exists.return_value = False
        self.local.setup.return_value = None
        self.local.create_repo.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_mkrepo(cmdp)
        self.assertEqual(status, 0)

    # def test_08__search_print_lines(self):
    #     """Test08 UdockerCLI()._search_print_lines()."""

    # @patch('udocker.cli.DockerIoAPI.search_get_page')
    # @patch('udocker.cli.HostInfo.termsize')
    # def test_09__search_repositories(self, mock_termsz, mock_doiasearch):
    #     """Test09 UdockerCLI()._search_repositories()."""
    #     repo_list = [{"count": 1, "next": "", "previous": "",
    #                   "results": [
    #                       {
    #                           "repo_name": "lipcomputing/ipyrad",
    #                           "short_description": "Docker to run ipyrad",
    #                           "star_count": 0,
    #                           "pull_count": 188,
    #                           "repo_owner": "",
    #                           "is_automated": True,
    #                           "is_official": False
    #                       }]}]
    #     mock_termsz.return_value = (40, "")
    #     mock_doiasearch.return_value = repo_list
    #     udoc = UdockerCLI(self.local)
    #     status = udoc._search_repositories("ipyrad")
    #     self.assertEqual(status, 0)

    @patch('udocker.cli.DockerIoAPI.get_tags')
    def test_10__list_tags(self, mock_gettags):
        """Test10 UdockerCLI()._list_tags()."""
        mock_gettags.return_value = ["t1"]
        udoc = UdockerCLI(self.local)
        status = udoc._list_tags("t1")
        self.assertEqual(status, 0)

        mock_gettags.return_value = None
        udoc = UdockerCLI(self.local)
        status = udoc._list_tags("t1")
        self.assertEqual(status, 1)

    @patch('udocker.cli.KeyStore.get')
    @patch('udocker.cli.DockerIoAPI.set_v2_login_token')
    @patch('udocker.cli.DockerIoAPI.search_init')
    @patch.object(UdockerCLI, '_search_repositories')
    @patch.object(UdockerCLI, '_list_tags')
    @patch.object(UdockerCLI, '_split_imagespec')
    @patch.object(UdockerCLI, '_set_repository')
    def test_11_do_search(self, mock_setrepo, mock_split, mock_listtags,
                          mock_searchrepo, mock_doiasearch, mock_doiasetv2,
                          mock_ksget):
        """Test11 UdockerCLI().do_search()."""
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_search(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "search", "--list-tags", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_setrepo.return_value = None
        mock_split.return_value = ("d1", "d2", "ipyrad", "d3")
        mock_doiasearch.return_value = None
        mock_ksget.return_value = "v2token1"
        mock_doiasetv2.return_value = None
        mock_listtags.return_value = ["t1", "t2"]
        udoc = UdockerCLI(self.local)
        status = udoc.do_search(cmdp)
        self.assertEqual(status, ["t1", "t2"])
        self.assertTrue(mock_setrepo.called)
        self.assertTrue(mock_doiasearch.called)
        self.assertTrue(mock_ksget.called)
        self.assertTrue(mock_doiasetv2.called)
        self.assertTrue(mock_listtags.called)

        argv = ["udocker", "search", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_setrepo.return_value = None
        mock_split.return_value = ("d1", "d2", "ipyrad", "d3")
        mock_doiasearch.return_value = None
        mock_ksget.return_value = "v2token1"
        mock_doiasetv2.return_value = None
        mock_searchrepo.return_value = 0
        udoc = UdockerCLI(self.local)
        status = udoc.do_search(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(mock_searchrepo.called)

    @patch('udocker.cli.Msg')
    @patch('udocker.cli.LocalFileAPI.load')
    @patch.object(UdockerCLI, '_check_imagerepo')
    def test_12_do_load(self, mock_chkimg, mock_load, mock_msg):
        """Test12 UdockerCLI().do_load()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_load(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "load", "-i", "ipyrad", "ipyimg"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_load(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(mock_load.called)

        argv = ["udocker", "load", "-i", "ipyrad", "ipyimg"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = True
        mock_load.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_load(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_load.called)

        argv = ["udocker", "load", "-i", "ipyrad", "ipyimg"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = True
        mock_load.return_value = ['docker-repo1', 'docker-repo2']
        udoc = UdockerCLI(self.local)
        status = udoc.do_load(cmdp)
        self.assertEqual(status, 0)

    @patch('udocker.cli.Msg')
    @patch('udocker.cli.os.path.exists')
    @patch('udocker.cli.LocalFileAPI.save')
    @patch.object(UdockerCLI, '_check_imagespec')
    def test_13_do_save(self, mock_chkimg, mock_save, mock_exists, mock_msg):
        """Test13 UdockerCLI().do_save()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_save(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "save", "-o", "ipyrad", "ipyimg:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_exists.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_save(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_exists.called)
        self.assertFalse(mock_chkimg.called)
        self.assertFalse(mock_save.called)

        argv = ["udocker", "save", "-o", "ipyrad", "ipyimg:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_exists.return_value = False
        mock_chkimg.return_value = ("ipyimg", "latest")
        mock_save.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_save(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_save.called)

        argv = ["udocker", "save", "-o", "ipyrad", "ipyimg:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_exists.return_value = False
        mock_chkimg.return_value = ("ipyimg", "latest")
        mock_save.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_save(cmdp)
        self.assertTrue(mock_exists.called)
        self.assertTrue(mock_chkimg.called)
        self.assertTrue(mock_save.called)
        self.assertEqual(status, 0)

    @patch('udocker.cli.LocalFileAPI.import_toimage')
    @patch('udocker.cli.LocalFileAPI.import_tocontainer')
    @patch('udocker.cli.LocalFileAPI.import_clone')
    @patch('udocker.cli.Msg')
    @patch.object(UdockerCLI, '_check_imagespec')
    def test_14_do_import(self, mock_chkimg, mock_msg, mock_impclone,
                          mock_impcont, mock_impimg):
        """Test14 UdockerCLI().do_import()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_import(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(mock_chkimg.called)
        self.assertFalse(mock_impclone.called)
        self.assertFalse(mock_impcont.called)
        self.assertFalse(mock_impimg.called)

        argv = ["udocker", "import", "ipyrad.tar", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_import(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_chkimg.called)
        self.assertFalse(mock_impimg.called)

        argv = ["udocker", "import", "ipyrad.tar", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_import(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(mock_chkimg.called)
        self.assertTrue(mock_impimg.called)

        argv = ["udocker", "import", "--clone", "ipyrad.tar", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        mock_impclone.return_value = "12345"
        udoc = UdockerCLI(self.local)
        status = udoc.do_import(cmdp)
        self.assertEqual(status, 0)
        self.assertFalse(mock_impcont.called)
        self.assertTrue(mock_impclone.called)

        argv = ["udocker", "import", "--tocontainer",
                "ipyrad.tar", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        mock_impcont.return_value = "12345"
        udoc = UdockerCLI(self.local)
        status = udoc.do_import(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(mock_impcont.called)

    @patch('udocker.cli.Msg')
    @patch('udocker.cli.ContainerStructure')
    def test_15_do_export(self, mock_cs, mock_msg):
        """Test15 UdockerCLI().do_export()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_export(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "export", "-o", "ipyrad.tar", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = ""
        udoc = UdockerCLI(self.local)
        status = udoc.do_export(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "export", "-o", "ipyrad.tar", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_cs.return_value.export_tofile.return_value = False
        self.local.get_container_id.return_value = "12345"
        udoc = UdockerCLI(self.local)
        status = udoc.do_export(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_cs.called)
        self.assertTrue(self.local.get_container_id.called)

        argv = ["udocker", "export", "-o", "ipyrad.tar", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_cs.return_value.export_tofile.return_value = True
        self.local.get_container_id.return_value = "12345"
        udoc = UdockerCLI(self.local)
        status = udoc.do_export(cmdp)
        self.assertEqual(status, 0)

        argv = ["udocker", "export",
                "--clone", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_cs.return_value.clone_tofile.return_value = True
        self.local.get_container_id.return_value = "12345"
        udoc = UdockerCLI(self.local)
        status = udoc.do_export(cmdp)
        self.assertEqual(status, 0)

    @patch('udocker.cli.LocalFileAPI.clone_container')
    @patch('udocker.cli.Msg')
    def test_16_do_clone(self, mock_msg, mock_clone):
        """Test16 UdockerCLI().do_clone()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_clone(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "clone", "ipyradcont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = ""
        udoc = UdockerCLI(self.local)
        status = udoc.do_clone(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(mock_clone.called)
        self.assertTrue(self.local.get_container_id.called)

        argv = ["udocker", "clone", "ipyradcont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = "12345"
        mock_clone.return_value = "54321"
        udoc = UdockerCLI(self.local)
        status = udoc.do_clone(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(mock_clone.called)

    @patch('udocker.cli.Msg')
    @patch('udocker.cli.KeyStore.put')
    @patch('udocker.cli.DockerIoAPI.get_v2_login_token')
    @patch.object(UdockerCLI, '_set_repository')
    def test_17_do_login(self, mock_setrepo, mock_dioalog,
                         mock_ksput, mock_msg):
        """Test17 UdockerCLI().do_login()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_login(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "login", "--username", "u1",
                "--password", "xx"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_setrepo.return_value = True
        mock_dioalog.return_value = "zx1"
        mock_ksput.return_value = 1
        udoc = UdockerCLI(self.local)
        status = udoc.do_login(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_setrepo.called)
        self.assertTrue(mock_dioalog.called)
        self.assertTrue(mock_ksput.called)

        argv = ["udocker", "login", "--username", "u1",
                "--password", "xx"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_setrepo.return_value = None
        mock_dioalog.return_value = "zx1"
        mock_ksput.return_value = 0
        udoc = UdockerCLI(self.local)
        status = udoc.do_login(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(mock_setrepo.called)
        self.assertTrue(mock_dioalog.called)
        self.assertTrue(mock_ksput.called)

    @patch('udocker.cli.Msg')
    @patch('udocker.cli.KeyStore')
    @patch.object(UdockerCLI, '_set_repository')
    def test_18_do_logout(self, mock_setrepo, mock_ks, mock_msg):
        """Test18 UdockerCLI().do_logout()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_logout(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "logout", "-a"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_setrepo.return_value = None
        mock_ks.return_value.erase.return_value = 0
        udoc = UdockerCLI(self.local)
        status = udoc.do_logout(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(mock_setrepo.called)
        self.assertTrue(mock_ks.return_value.erase.called)

        argv = ["udocker", "logout"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_setrepo.return_value = None
        mock_ks.return_value.delete.return_value = 1
        udoc = UdockerCLI(self.local)
        status = udoc.do_logout(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_setrepo.called)
        self.assertTrue(mock_ks.return_value.delete.called)

    @patch.object(UdockerCLI, '_set_repository')
    @patch.object(UdockerCLI, '_check_imagespec')
    @patch('udocker.cli.DockerIoAPI')
    @patch('udocker.cli.KeyStore.get')
    @patch('udocker.cli.Msg')
    def test_19_do_pull(self, mock_msg, mock_ksget, mock_dioa,
                        mock_chkimg, mock_setrepo):
        """Test19 UdockerCLI().do_pull()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_pull(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "pull", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        mock_setrepo.return_value = None
        mock_ksget.return_value = "zx1"
        mock_dioa.return_value.set_v2_login_token.return_value = None
        mock_dioa.return_value.get.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_pull(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_chkimg.called)
        self.assertTrue(mock_setrepo.called)
        self.assertTrue(mock_ksget.called)
        self.assertTrue(mock_dioa.return_value.set_v2_login_token.called)
        self.assertTrue(mock_dioa.return_value.get.called)

        argv = ["udocker", "pull", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        mock_setrepo.return_value = None
        mock_ksget.return_value = "zx1"
        mock_dioa.return_value.set_v2_login_token.return_value = None
        mock_dioa.return_value.get.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_pull(cmdp)
        self.assertEqual(status, 0)

    @patch.object(UdockerCLI, '_check_imagespec')
    @patch('udocker.cli.ContainerStructure')
    @patch('udocker.cli.DockerIoAPI')
    @patch('udocker.cli.Msg')
    def test_20__create(self, mock_msg, mock_dioapi,
                        mock_cstruct, mock_chkimg):
        """Test20 UdockerCLI()._create()."""
        mock_msg.level = 0
        mock_dioapi.return_value.is_repo_name.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc._create("IMAGE:TAG")
        self.assertFalse(status)
        self.assertTrue(mock_msg.return_value.err.called)

        mock_dioapi.return_value.is_repo_name.return_value = True
        mock_chkimg.return_value = ("", "TAG")
        mock_cstruct.return_value.create.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc._create("IMAGE:TAG")
        self.assertFalse(status)

        mock_dioapi.return_value.is_repo_name.return_value = True
        mock_chkimg.return_value = ("IMAGE", "TAG")
        mock_cstruct.return_value.create.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc._create("IMAGE:TAG")
        self.assertTrue(status)

    @patch.object(UdockerCLI, '_create')
    @patch('udocker.cli.Msg')
    def test_21_do_create(self, mock_msg, mock_create):
        """Test21 UdockerCLI().do_create()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_create(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(mock_create.called)

        argv = ["udocker", "create", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_create.return_value = ""
        udoc = UdockerCLI(self.local)
        status = udoc.do_create(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_create.called)

        argv = ["udocker", "create", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_create.return_value = "12345"
        udoc = UdockerCLI(self.local)
        status = udoc.do_create(cmdp)
        self.assertEqual(status, 0)
        self.assertFalse(self.local.set_container_name.called)

        argv = ["udocker", "create", "--name=mycont", "ipyrad:latest"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_create.return_value = "12345"
        self.local.set_container_name.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_create(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.set_container_name.called)

    # def test_22__get_run_options(self):
    #    """Test22 UdockerCLI()._get_run_options()"""

    @patch('udocker.cli.ExecutionMode')
    @patch('udocker.cli.Msg')
    @patch.object(UdockerCLI, 'do_pull')
    @patch.object(UdockerCLI, '_create')
    @patch.object(UdockerCLI, '_check_imagespec')
    def test_23_do_run(self, mock_chkimg, mock_create, mock_pull,
                       mock_msg, mock_exec):
        """Test23 UdockerCLI().do_run()."""
        mock_msg.level = 0
        mock_pull.return_value = None
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_run(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "run"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_pull.return_value = None
        udoc = UdockerCLI(self.local)
        status = udoc.do_run(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "run", "--location=/tmp/udocker", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_pull.return_value = None
        mock_exec.return_value.get_engine.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_run(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_exec.return_value.get_engine.called)

        mock_pull.return_value = None
        exeng_patch = patch("udocker.engine.proot.PRootEngine")
        proot = exeng_patch.start()
        mock_proot = Mock()
        proot.return_value = mock_proot

        argv = ["udocker", "run", "--location=/tmp/udocker", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_pull.return_value = None
        mock_exec.return_value.get_engine.return_value = proot
        proot.run.return_value = 0
        udoc = UdockerCLI(self.local)
        status = udoc.do_run(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(proot.run.called)
        self.assertFalse(self.local.isprotected_container.called)

        argv = ["udocker", "run", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = ""
        mock_pull.return_value = None
        mock_exec.return_value.get_engine.return_value = proot
        proot.run.return_value = 0
        mock_chkimg.return_value = ("ipyrad", "latest")
        self.local.cd_imagerepo.return_value = True
        mock_create.return_value = "12345"
        udoc = UdockerCLI(self.local)
        status = udoc.do_run(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(mock_chkimg.called)
        self.assertTrue(self.local.cd_imagerepo.called)
        self.assertTrue(mock_create.called)

        exeng_patch.stop()

    def test_24_do_images(self):
        """Test24 UdockerCLI().do_images()."""
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_images(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "images", "-l"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_imagerepos.return_value = [("img1", "tag1")]
        self.local.isprotected_imagerepo.return_value = False
        self.local.cd_imagerepo.return_value = "/img1"
        self.local.get_layers.return_value = [("l1", 1024)]
        udoc = UdockerCLI(self.local)
        status = udoc.do_images(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(self.local.get_imagerepos.called)
        self.assertTrue(self.local.isprotected_imagerepo.called)
        self.assertTrue(self.local.cd_imagerepo.called)
        self.assertTrue(self.local.get_layers.called)

    @patch('udocker.cli.ExecutionMode')
    def test_25_do_ps(self, mock_exec):
        """Test25 UdockerCLI().do_ps()."""
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_ps(cmdp)
        self.assertEqual(status, 1)

        exeng_patch = patch("udocker.engine.proot.PRootEngine")
        proot = exeng_patch.start()
        mock_proot = Mock()
        proot.return_value = mock_proot
        cdir = "/home/u1/.udocker/containers"
        argv = ["udocker", "ps", "-m", "-s"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_containers_list.return_value = [[cdir, "/", "a"]]
        mock_exec.return_value.get_engine.return_value = proot
        self.local.isprotected_container.return_value = False
        self.local.iswriteable_container.return_value = True
        self.local.get_size.return_value = 1024
        udoc = UdockerCLI(self.local)
        status = udoc.do_ps(cmdp)
        self.assertEqual(status, 0)
        exeng_patch.stop()

    @patch('udocker.cli.Msg')
    def test_26_do_rm(self, mock_msg):
        """Test26 UdockerCLI().do_rm()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_rm(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "rm"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_rm(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(self.local.get_container_id.called)

        argv = ["udocker", "rm", "mycont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = None
        udoc = UdockerCLI(self.local)
        status = udoc.do_rm(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.get_container_id.called)
        self.assertFalse(self.local.isprotected_container.called)

        argv = ["udocker", "rm", "mycont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = "12345"
        self.local.isprotected_container.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_rm(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.isprotected_container.called)
        self.assertFalse(self.local.del_container.called)

        argv = ["udocker", "rm", "mycont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = "12345"
        self.local.isprotected_container.return_value = False
        self.local.del_container.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_rm(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.del_container.called)

        argv = ["udocker", "rm", "mycont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = "12345"
        self.local.isprotected_container.return_value = False
        self.local.del_container.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_rm(cmdp)
        self.assertEqual(status, 0)

    @patch('udocker.cli.Msg')
    @patch.object(UdockerCLI, '_check_imagespec')
    def test_27_do_rmi(self, mock_chkimg, mock_msg):
        """Test27 UdockerCLI().do_rmi()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_rmi(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "rmi"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_rmi(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(mock_chkimg.called)
        self.assertFalse(self.local.isprotected_imagerepo.called)

        argv = ["udocker", "rmi", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        self.local.isprotected_imagerepo.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_rmi(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.isprotected_imagerepo.called)
        self.assertFalse(self.local.del_imagerepo.called)

        argv = ["udocker", "rmi", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        self.local.isprotected_imagerepo.return_value = False
        self.local.del_imagerepo.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_rmi(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.del_imagerepo.called)

        argv = ["udocker", "rmi", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        self.local.isprotected_imagerepo.return_value = False
        self.local.del_imagerepo.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_rmi(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(self.local.del_imagerepo.called)

    @patch('udocker.cli.Msg')
    @patch.object(UdockerCLI, '_check_imagespec')
    def test_28_do_protect(self, mock_chkimg, mock_msg):
        """Test28 UdockerCLI().do_protect()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_protect(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(self.local.get_container_id.called)

        argv = ["udocker", "protect"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = False
        mock_chkimg.return_value = ("", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_protect(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(mock_chkimg.called)
        self.assertFalse(self.local.protect_container.called)
        self.assertFalse(self.local.protect_imagerepo.called)

        argv = ["udocker", "protect", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = False
        mock_chkimg.return_value = ("", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_protect(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "protect", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = False
        self.local.protect_imagerepo.return_value = True
        mock_chkimg.return_value = ("ipyrad", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_protect(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(self.local.protect_imagerepo.called)

        argv = ["udocker", "protect", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = True
        self.local.protect_container.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_protect(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(self.local.protect_container.called)

        argv = ["udocker", "protect", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = True
        self.local.protect_container.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_protect(cmdp)
        self.assertEqual(status, 1)

    @patch('udocker.cli.Msg')
    @patch.object(UdockerCLI, '_check_imagespec')
    def test_29_do_unprotect(self, mock_chkimg, mock_msg):
        """Test29 UdockerCLI().do_unprotect()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_unprotect(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(self.local.get_container_id.called)

        argv = ["udocker", "protect"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = False
        mock_chkimg.return_value = ("", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_unprotect(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(mock_chkimg.called)
        self.assertFalse(self.local.unprotect_container.called)
        self.assertFalse(self.local.unprotect_imagerepo.called)

        argv = ["udocker", "protect", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = False
        mock_chkimg.return_value = ("", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_unprotect(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "protect", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = False
        self.local.unprotect_imagerepo.return_value = True
        mock_chkimg.return_value = ("ipyrad", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_unprotect(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(self.local.unprotect_imagerepo.called)

        argv = ["udocker", "protect", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = True
        self.local.unprotect_container.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_unprotect(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(self.local.get_container_id.called)
        self.assertTrue(self.local.unprotect_container.called)

        argv = ["udocker", "protect", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = True
        self.local.unprotect_container.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_unprotect(cmdp)
        self.assertEqual(status, 1)

    @patch('udocker.cli.Msg')
    def test_30_do_name(self, mock_msg):
        """Test30 UdockerCLI().do_name()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_name(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "name"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_name(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.get_container_id.called)
        self.assertFalse(self.local.set_container_name.called)

        argv = ["udocker", "name", "12345", "mycont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = True
        self.local.set_container_name.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_name(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.set_container_name.called)

        argv = ["udocker", "name", "12345", "mycont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = True
        self.local.set_container_name.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_name(cmdp)
        self.assertEqual(status, 0)

    @patch('udocker.cli.Msg')
    def test_31_do_rename(self, mock_msg):
        """Test31 UdockerCLI().do_rename()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_rename(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "rename", "contname", "newname"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.side_effect = ["", ""]
        udoc = UdockerCLI(self.local)
        status = udoc.do_rename(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.get_container_id.call_count, 1)

        argv = ["udocker", "rename", "contname", "newname"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.side_effect = ["123", "543"]
        udoc = UdockerCLI(self.local)
        status = udoc.do_rename(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.get_container_id.call_count, 2)

        argv = ["udocker", "rename", "contname", "newname"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.side_effect = ["123", ""]
        self.local.del_container_name.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_rename(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.del_container_name.called)
        self.assertFalse(self.local.set_container_name.called)

        argv = ["udocker", "rename", "contname", "newname"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.side_effect = ["123", ""]
        self.local.del_container_name.return_value = True
        self.local.set_container_name.side_effect = [False, True]
        udoc = UdockerCLI(self.local)
        status = udoc.do_rename(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.set_container_name.call_count, 2)

        argv = ["udocker", "rename", "contname", "newname"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.side_effect = ["123", ""]
        self.local.del_container_name.return_value = True
        self.local.set_container_name.side_effect = [True, True]
        udoc = UdockerCLI(self.local)
        status = udoc.do_rename(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(self.local.set_container_name.call_count, 1)

    @patch('udocker.cli.Msg')
    def test_32_do_rmname(self, mock_msg):
        """Test32 UdockerCLI().do_rmname()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_rmname(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "rmname"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_rmname(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(self.local.del_container_name.called)

        argv = ["udocker", "rmname", "contname"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.del_container_name.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_rmname(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.del_container_name.called)

        argv = ["udocker", "rmname", "contname"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.del_container_name.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_rmname(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(self.local.del_container_name.called)

    @patch.object(UdockerCLI, '_check_imagespec')
    @patch('udocker.cli.json.dumps')
    @patch('udocker.cli.ContainerStructure.get_container_attr')
    @patch('udocker.cli.Msg')
    def test_33_do_inspect(self, mock_msg, mock_csattr, mock_jdump,
                           mock_chkimg):
        """Test33 UdockerCLI().do_inspect()."""
        cont_insp = \
                {
                    "architecture": "amd64",
                    "config": {
                        "AttachStderr": False,
                        "AttachStdin": False,
                        "AttachStdout": False,
                        "Cmd": [
                            "/bin/bash"
                        ],
                        "Domainname": "",
                        "Entrypoint": None,
                        "Env": [
                            "PATH=/usr/local/sbin"
                        ],
                        "Hostname": "",
                        "Image": "sha256:05725a",
                        "Labels": {
                            "org.opencontainers.image.vendor": "CentOS"
                        },
                        "WorkingDir": ""
                    },
                    "container": "c171c",
                    "container_config": {
                        "ArgsEscaped": True,
                        "Cmd": ["/bin/sh", "-c"],
                        "Domainname": "",
                        "Env": [
                            "PATH=/usr/local/sbin"
                        ],
                        "Hostname": "c171c5a1528a",
                        "Image": "sha256:05725a",
                        "Labels": {
                            "org.label-schema.license": "GPLv2",
                            "org.label-schema.name": "CentOS Base Image",
                            "org.opencontainers.image.vendor": "CentOS"
                        },
                        "WorkingDir": ""
                    },
                    "created": "2020-05-05T21",
                    "docker_version": "18.09.7",
                    "id": "e72c1",
                    "os": "linux",
                    "parent": "61dc7"
                }

        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = ""
        udoc = UdockerCLI(self.local)
        status = udoc.do_inspect(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "inspect"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = ""
        mock_chkimg.return_value = ("", "latest")
        self.local.cd_imagerepo.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_inspect(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "inspect"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = ""
        mock_chkimg.return_value = ("ipyrad", "latest")
        self.local.cd_imagerepo.return_value = True
        self.local.get_image_attributes.return_value = (cont_insp, "")
        mock_jdump.return_value = cont_insp
        udoc = UdockerCLI(self.local)
        status = udoc.do_inspect(cmdp)
        self.assertEqual(status, 0)

        argv = ["udocker", "inspect", "-p", "123"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.get_container_id.return_value = "123"
        mock_chkimg.return_value = ("ipyrad", "latest")
        self.local.cd_imagerepo.return_value = True
        self.local.get_image_attributes.return_value = (cont_insp, "")
        mock_csattr.return_value = ("/ROOT/cont", cont_insp)
        mock_jdump.return_value = cont_insp
        udoc = UdockerCLI(self.local)
        status = udoc.do_inspect(cmdp)
        self.assertEqual(status, 0)

    @patch.object(UdockerCLI, '_check_imagespec')
    @patch('udocker.cli.Msg')
    def test_34_do_verify(self, mock_msg, mock_chkimg):
        """Test34 UdockerCLI().do_verify()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        udoc = UdockerCLI(self.local)
        status = udoc.do_verify(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "verify", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        self.local.cd_imagerepo.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_verify(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "verify", "ipyrad"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_chkimg.return_value = ("ipyrad", "latest")
        self.local.cd_imagerepo.return_value = True
        self.local.verify_image.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_verify(cmdp)
        self.assertEqual(status, 0)

    @patch('udocker.cli.ExecutionMode')
    @patch('udocker.cli.NvidiaMode')
    @patch('udocker.cli.FileUtil.rchmod')
    @patch('udocker.cli.Unshare.namespace_exec')
    @patch('udocker.cli.MountPoint')
    @patch('udocker.cli.FileBind')
    @patch('udocker.cli.Msg')
    def test_35_do_setup(self, mock_msg, mock_fb, mock_mp,
                         mock_unshr, mock_furchmod, mock_nv, mock_execm):
        """Test35 UdockerCLI().do_setup()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_setup(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "setup"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.cd_container.return_value = ""
        udoc = UdockerCLI(self.local)
        status = udoc.do_setup(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.cd_container.called)

        argv = ["udocker", "setup", "--execmode=P2", "mycont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.cd_container.return_value = "/ROOT/cont1"
        self.local.isprotected_container.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_setup(cmdp)
        self.assertEqual(status, 1)
        self.assertTrue(self.local.isprotected_container.called)

        argv = ["udocker", "setup", "--execmode=P2",
                "--purge", "--fixperm", "--nvidia", "mycont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.cd_container.return_value = "/ROOT/cont1"
        self.local.isprotected_container.return_value = False
        mock_msg.level = 0
        mock_fb.return_value.restore.return_value = None
        mock_mp.return_value.restore.return_value = None
        mock_unshr.return_value = None
        mock_furchmod.return_value = None
        mock_nv.return_value.set_mode.return_value = None
        mock_execm.return_value.set_mode.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_setup(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(mock_fb.return_value.restore.called)
        self.assertTrue(mock_mp.return_value.restore.called)
        self.assertTrue(mock_unshr.called)
        self.assertTrue(mock_furchmod.called)
        self.assertTrue(mock_nv.return_value.set_mode.called)
        self.assertTrue(mock_execm.return_value.set_mode.called)

        argv = ["udocker", "setup", "--execmode=P2",
                "--purge", "--fixperm", "--nvidia", "mycont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.cd_container.return_value = "/ROOT/cont1"
        self.local.isprotected_container.return_value = False
        mock_msg.level = 0
        mock_fb.return_value.restore.return_value = None
        mock_mp.return_value.restore.return_value = None
        mock_unshr.return_value = None
        mock_furchmod.return_value = None
        mock_nv.return_value.set_mode.return_value = None
        mock_execm.return_value.set_mode.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_setup(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "setup", "mycont"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        self.local.cd_container.return_value = "/ROOT/cont1"
        self.local.isprotected_container.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_setup(cmdp)
        self.assertEqual(status, 0)

    @patch('udocker.cli.UdockerTools')
    @patch('udocker.cli.Msg')
    def test_36_do_install(self, mock_msg, mock_utools):
        """Test36 UdockerCLI().do_install()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_install(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "install", "--force", "--purge"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_utools.return_value.purge.return_value = None
        mock_utools.return_value.install.return_value = False
        udoc = UdockerCLI(self.local)
        status = udoc.do_install(cmdp)
        self.assertEqual(status, 1)

        argv = ["udocker", "install", "--force", "--purge"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        mock_utools.return_value.purge.return_value = None
        mock_utools.return_value.install.return_value = True
        udoc = UdockerCLI(self.local)
        status = udoc.do_install(cmdp)
        self.assertEqual(status, 0)

    @patch('udocker.cli.Msg')
    def test_37_do_showconf(self, mock_msg):
        """Test37 UdockerCLI().do_showconf()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_showconf(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(mock_msg.return_value.out.called)

        argv = ["udocker", "showconf"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_showconf(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(mock_msg.return_value.out.called)

    @patch('udocker.cli.Msg')
    def test_38_do_version(self, mock_msg):
        """Test38 UdockerCLI().do_version()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_version(cmdp)
        self.assertEqual(status, 1)
        self.assertFalse(mock_msg.return_value.out.called)

        argv = ["udocker", "version"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_version(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(mock_msg.return_value.out.called)

    @patch('udocker.cli.Msg')
    def test_39_do_help(self, mock_msg):
        """Test39 UdockerCLI().do_help()."""
        mock_msg.level = 0
        argv = ["udocker", "-h"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        udoc = UdockerCLI(self.local)
        status = udoc.do_help(cmdp)
        self.assertEqual(status, 0)
        self.assertTrue(mock_msg.return_value.out.called)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Docker API integration"""

import os
import sys
import re
import base64
import json

from udocker.config import Config
from udocker.msg import Msg
from udocker.commonlocalfile import CommonLocalFileApi
from udocker.utils.fileutil import FileUtil
from udocker.utils.curl import GetURL
from udocker.utils.chksum import ChkSUM
from udocker.helper.unique import Unique


class DockerIoAPI(object):
    """Class to encapsulate the access to the Docker Hub service
    Allows to search and download images from Docker Hub
    """

    def __init__(self, localrepo):
        self.index_url = Config.conf['dockerio_index_url']
        self.registry_url = Config.conf['dockerio_registry_url']
        self.v1_auth_header = ""
        self.v2_auth_header = ""
        self.v2_auth_token = ""
        self.localrepo = localrepo
        self.curl = GetURL()
        self.search_pause = True
        self.search_page = 0
        self.search_ended = False

    def set_proxy(self, http_proxy):
        """Select a socks http proxy for API access and file download"""
        self.curl.set_proxy(http_proxy)

    def set_registry(self, registry_url):
        """Change docker registry url"""
        self.registry_url = registry_url

    def set_index(self, index_url):
        """Change docker index url"""
        self.index_url = index_url

    def is_repo_name(self, imagerepo):
        """Check if name matches authorized characters for a docker repo"""
        if imagerepo and re.match("^[a-zA-Z0-9][a-zA-Z0-9-_./:]+$", imagerepo):
            return True
        return False

    def _get_url(self, *args, **kwargs):
        """Encapsulates the call to GetURL.get() so that authentication
        for v1 and v2 repositories can be treated differently.
        Example:
             _get_url(url, ctimeout=5, timeout=5, header=[]):
        """
        url = str(args[0])
        if "RETRY" not in kwargs:
            kwargs["RETRY"] = 3
        if "FOLLOW" not in kwargs:
            kwargs["FOLLOW"] = 3
        kwargs["RETRY"] -= 1
        (hdr, buf) = self.curl.get(*args, **kwargs)
        Msg().out(f"Info: header: {hdr.data}", l=Msg.DBG)
        Msg().out(f"Info: buffer: {buf.getvalue()}", l=Msg.DBG)
        status_code = self.curl.get_status_code(hdr.data["X-ND-HTTPSTATUS"])
        if status_code == 200:
            return (hdr, buf)
        if not kwargs["RETRY"]:
            hdr.data["X-ND-CURLSTATUS"] = 13  # Permission denied
            return (hdr, buf)
        auth_kwargs = kwargs.copy()
        if "location" not in hdr.data:
            kwargs["FOLLOW"] = 3
        if "location" in hdr.data and hdr.data['location']:
            if not kwargs["FOLLOW"]:
                hdr.data["X-ND-CURLSTATUS"] = 13
                return (hdr, buf)
            kwargs["FOLLOW"] -= 1
            args = [hdr.data['location']]
            if "header" in auth_kwargs:
                del auth_kwargs["header"]
        elif status_code == 401:
            if "www-authenticate" in hdr.data:
                www_authenticate = hdr.data["www-authenticate"]
                if not "realm" in www_authenticate:
                    return (hdr, buf)
                if 'error="insufficient_scope"' in www_authenticate:
                    return (hdr, buf)
                auth_header = ""
                if "/v2/" in url:
                    auth_header = self._get_v2_auth(www_authenticate,
                                                    kwargs["RETRY"])
                if "/v1/" in url:
                    auth_header = self._get_v1_auth(www_authenticate)
                auth_kwargs.update({"header": [auth_header]})
        (hdr, buf) = self._get_url(*args, **auth_kwargs)
        return (hdr, buf)

    def _get_file(self, url, filename, cache_mode):
        """Get a file and check its size. Optionally enable other
        capabilities such as caching to check if the
        file already exists locally and whether its size is the
        same to avoid downloaded it again.
        """
        match = re.search("/([^/:]+):(\\S+)$", filename)
        if match:
            layer_f_chksum = ChkSUM().hash(filename, match.group(1))
            if layer_f_chksum == match.group(2):
                return True             # is cached skip download
            cache_mode = 0
        if self.curl.cache_support and cache_mode:
            if cache_mode == 1:
                (hdr, dummy) = self._get_url(url, nobody=1)
            elif cache_mode == 3:
                (hdr, dummy) = self._get_url(url, sizeonly=True)
            remote_size = self.curl.get_content_length(hdr)
            if remote_size == FileUtil(filename).size():
                return True             # is cached skip download
        else:
            remote_size = -1
        resume = False
        if filename.endswith("layer"):
            resume = True
        (hdr, dummy) = self._get_url(url, ofile=filename, resume=resume)
        if self.curl.get_status_code(hdr.data["X-ND-HTTPSTATUS"]) != 200:
            return False
        if remote_size == -1:
            remote_size = self.curl.get_content_length(hdr)
        if (remote_size != FileUtil(filename).size() and
                hdr.data["X-ND-CURLSTATUS"]):
            Msg().err("Error: file size mismatch:", filename,
                      remote_size, FileUtil(filename).size())
            return False
        return True

    def _split_fields(self, buf):
        """Split  fields, used in the web authentication"""
        all_fields = {}
        for field in buf.split(','):
            pair = field.split('=', 1)
            if len(pair) == 2:
                all_fields[pair[0]] = pair[1].strip('"')
        return all_fields

    def is_v1(self):
        """Check if registry is of type v1"""
        for prefix in ("/v1", "/v1/_ping"):
            (hdr, dummy) = self._get_url(self.index_url + prefix)
            try:
                if ("200" in hdr.data["X-ND-HTTPSTATUS"] or
                        "401" in hdr.data["X-ND-HTTPSTATUS"]):
                    return True
            except (KeyError, AttributeError, TypeError):
                pass
        return False

    def has_search_v1(self, url=None):
        """Check if registry has search capabilities in v1"""
        if url is None:
            url = self.index_url
        (hdr, dummy) = self._get_url(url + "/v1/search")
        try:
            if ("200" in hdr.data["X-ND-HTTPSTATUS"] or
                    "401" in hdr.data["X-ND-HTTPSTATUS"]):
                return True
        except (KeyError, AttributeError, TypeError):
            pass
        return False

    def get_v1_repo(self, imagerepo):
        """Get list of images in a repo from Docker Hub"""
        url = self.index_url + "/v1/repositories/" + imagerepo + "/images"
        Msg().out("Info: repo url", url, l=Msg.DBG)
        (hdr, buf) = self._get_url(url, header=["X-Docker-Token: true"])
        try:
            self.v1_auth_header = "Authorization: Token " + \
                hdr.data["x-docker-token"]
            return hdr.data, json.loads(buf.getvalue())
        except (IOError, OSError, AttributeError,
                ValueError, TypeError, KeyError):
            self.v1_auth_header = ""
            return hdr.data, []

    def _get_v1_auth(self, www_authenticate):
        """Authentication for v1 API"""
        if "Token" in www_authenticate:
            return self.v1_auth_header
        return ""

    def get_v1_image_tags(self, imagerepo, tags_only=False):
        """Get list of tags in a repo from Docker Hub"""
        (hdr_data, dummy) = self.get_v1_repo(imagerepo)
        try:
            endpoint = "http://" + hdr_data["x-docker-endpoints"]
        except KeyError:
            endpoint = self.index_url
        url = endpoint + "/v1/repositories/" + imagerepo + "/tags"
        Msg().out("Info: tags url", url, l=Msg.DBG)
        (dummy, buf) = self._get_url(url)
        tags = []
        try:
            if tags_only:
                for tag in json.loads(buf.getvalue()):
                    tags.append(tag["name"])
                return tags

            return json.loads(buf.getvalue())
        except (IOError, OSError, AttributeError, ValueError, TypeError):
            return []

    def get_v1_image_tag(self, endpoint, imagerepo, tag):
        """Get list of tags in a repo from Docker Hub"""
        url = endpoint + "/v1/repositories/" + imagerepo + "/tags/" + tag
        Msg().out("Info: tags url", url, l=Msg.DBG)
        (hdr, buf) = self._get_url(url)
        try:
            return (hdr.data, json.loads(buf.getvalue()))
        except (IOError, OSError, AttributeError, ValueError, TypeError):
            return (hdr.data, [])

    def get_v1_image_ancestry(self, endpoint, image_id):
        """Get the ancestry which is an ordered list of layers"""
        url = endpoint + "/v1/images/" + image_id + "/ancestry"
        Msg().out("Info: ancestry url", url, l=Msg.DBG)
        (hdr, buf) = self._get_url(url)
        try:
            return (hdr.data, json.loads(buf.getvalue()))
        except (IOError, OSError, AttributeError, ValueError, TypeError):
            return (hdr.data, [])

    def get_v1_image_json(self, endpoint, layer_id):
        """Get the JSON metadata for a specific layer"""
        url = endpoint + "/v1/images/" + layer_id + "/json"
        Msg().out("Info: json url", url, l=Msg.DBG)
        filename = self.localrepo.layersdir + '/' + layer_id + ".json"
        if self._get_file(url, filename, 0):
            self.localrepo.add_image_layer(filename)
            return True
        return False

    def get_v1_image_layer(self, endpoint, layer_id):
        """Get a specific layer data file (layer files are tarballs)"""
        url = endpoint + "/v1/images/" + layer_id + "/layer"
        Msg().out("Info: layer url", url, l=Msg.DBG)
        filename = self.localrepo.layersdir + '/' + layer_id + ".layer"
        if self._get_file(url, filename, 3):
            self.localrepo.add_image_layer(filename)
            return True
        return False

    def get_v1_layers_all(self, endpoint, layer_list):
        """Using a layer list download data and metadata files"""
        files = []
        if layer_list:
            for layer_id in reversed(layer_list):
                Msg().out("Info: downloading layer", layer_id, l=Msg.INF)
                filesize = self.get_v1_image_json(endpoint, layer_id)
                if not filesize:
                    return []
                files.append(layer_id + ".json")
                filesize = self.get_v1_image_layer(endpoint, layer_id)
                if not filesize:
                    return []
                files.append(layer_id + ".layer")
        return files

    def _get_v2_auth(self, www_authenticate, retry):
        """Authentication for v2 API"""
        auth_header = ""
        (bearer, auth_data) = www_authenticate.rsplit(' ', 1)
        if bearer == "Bearer":
            auth_fields = self._split_fields(auth_data)
            if "realm" in auth_fields:
                auth_url = auth_fields["realm"] + '?'
                for (field, value) in auth_fields.items():
                    if field != "realm":
                        auth_url += field + '=' + value + '&'
                header = []
                if self.v2_auth_token:
                    header = [f"Authorization: Basic {self.v2_auth_token}"]
                (dummy, auth_buf) = self._get_url(auth_url, header=header,
                                                  RETRY=retry)
                if sys.version_info[0] >= 3:
                    token_buf = auth_buf.getvalue().decode()
                else:
                    token_buf = auth_buf.getvalue()
                if token_buf and "token" in token_buf:
                    try:
                        auth_token = json.loads(token_buf)
                    except (IOError, OSError, AttributeError,
                            ValueError, TypeError):
                        return auth_header
                    auth_header = "Authorization: Bearer " + \
                        auth_token["token"]
                    self.v2_auth_header = auth_header
        # PR #126
        elif 'BASIC' in bearer or 'Basic' in bearer:
            auth_header = f"Authorization: Basic {self.v2_auth_token}"
            self.v2_auth_header = auth_header
        return auth_header

    def get_v2_login_token(self, username, password):
        """Get a login token from username and password"""
        if not (username and password):
            return ""
        try:
            self.v2_auth_token = \
                base64.b64encode((f"{username}:{password}").encode("utf-8")).decode("ascii")
        except (KeyError, AttributeError, TypeError, ValueError, NameError):
            self.v2_auth_token = ""
        return self.v2_auth_token

    def set_v2_login_token(self, v2_auth_token):
        """Load previously created login token"""
        self.v2_auth_token = v2_auth_token

    def is_v2(self):
        """Check if registry is of type v2"""
        (hdr, dummy) = self._get_url(self.registry_url + "/v2/")
        try:
            if ("200" in hdr.data["X-ND-HTTPSTATUS"] or
                    "401" in hdr.data["X-ND-HTTPSTATUS"]):
                return True
        except (KeyError, AttributeError, TypeError):
            pass
        return False

    def has_search_v2(self, url=None):
        """Check if registry has search capabilities in v2"""
        if url is None:
            url = self.registry_url
        (hdr, dummy) = self._get_url(url + "/v2/search/repositories")
        try:
            if ("200" in hdr.data["X-ND-HTTPSTATUS"] or
                    "401" in hdr.data["X-ND-HTTPSTATUS"]):
                return True
        except (KeyError, AttributeError, TypeError):
            pass
        return False

    def get_v2_image_tags(self, imagerepo, tags_only=False):
        """Get list of tags in a repo from Docker Hub"""
        url = self.registry_url + "/v2/" + imagerepo + "/tags/list"
        Msg().out("Info: tags url", url, l=Msg.DBG)
        (dummy, buf) = self._get_url(url)
        tags = []
        try:
            if tags_only:
                for tag in json.loads(buf.getvalue())["tags"]:
                    tags.append(tag)
                return tags

            return json.loads(buf.getvalue())
        except (IOError, OSError, AttributeError, ValueError, TypeError):
            return []

    def get_v2_image_manifest(self, imagerepo, tag):
        """Get the image manifest which contains JSON metadata
        that is common to all layers in this image tag
        """
        url = self.registry_url + "/v2/" + imagerepo + \
            "/manifests/" + tag
        Msg().out("Info: manifest url", url, l=Msg.DBG)
        (hdr, buf) = self._get_url(url)
        try:
            return (hdr.data, json.loads(buf.getvalue()))
        except (IOError, OSError, AttributeError, ValueError, TypeError):
            return (hdr.data, [])

    def get_v2_image_layer(self, imagerepo, layer_id):
        """Get one image layer data file (tarball)"""
        url = self.registry_url + "/v2/" + imagerepo + \
            "/blobs/" + layer_id
        Msg().out("Info: layer url", url, l=Msg.DBG)
        filename = self.localrepo.layersdir + '/' + layer_id
        if self._get_file(url, filename, 3):
            self.localrepo.add_image_layer(filename)
            return True
        return False

    def get_v2_layers_all(self, imagerepo, fslayers):
        """Get all layer data files belonging to a image tag"""
        files = []
        blob = ""
        if fslayers:
            for layer in reversed(fslayers):
                if "blobSum" in layer:
                    blob = layer["blobSum"]
                elif "digest" in layer:
                    blob = layer["digest"]
                Msg().out("Info: downloading layer", blob, l=Msg.INF)
                if not self.get_v2_image_layer(imagerepo, blob):
                    return []
                files.append(blob)
        return files

    def get_v2(self, imagerepo, tag):
        """Pull container with v2 API"""
        files = []
        (hdr_data, manifest) = self.get_v2_image_manifest(imagerepo, tag)
        status = self.curl.get_status_code(hdr_data["X-ND-HTTPSTATUS"])
        if status == 401:
            Msg().err("Error: manifest not found or not authorized")
            return []
        if status != 200:
            Msg().err("Error: pulling manifest:")
            return []
        try:
            if not (self.localrepo.setup_tag(tag) and
                    self.localrepo.set_version("v2")):
                Msg().err("Error: setting localrepo v2 tag and version")
                return []
            self.localrepo.save_json("manifest", manifest)
            Msg().out(f"Info: v2 layers: {imagerepo}", l=Msg.DBG)
            if "fsLayers" in manifest:
                files = self.get_v2_layers_all(imagerepo,
                                               manifest["fsLayers"])
            elif "layers" in manifest:
                if "config" in manifest:
                    manifest["layers"].append(manifest["config"])
                files = self.get_v2_layers_all(imagerepo,
                                               manifest["layers"])
            else:
                Msg().err("Error: layers section missing in manifest")
        except (KeyError, AttributeError, IndexError, ValueError, TypeError):
            pass
        return files

    def _get_v1_id_from_tags(self, tags_obj, tag):
        """Get image id from array of tags"""
        if isinstance(tags_obj, dict):
            try:
                return tags_obj[tag]
            except KeyError:
                pass
        elif isinstance(tags_obj, list):
            try:
                for tag_dict in tags_obj:
                    if tag_dict["name"] == tag:
                        return tag_dict["layer"]
            except KeyError:
                pass
        return ""

    def _get_v1_id_from_images(self, images_array, short_id):
        """Get long image id from array of images using the short id"""
        try:
            for image_dict in images_array:
                if image_dict["id"][0:8] == short_id:
                    return image_dict["id"]
        except KeyError:
            pass
        return ""

    def get_v1(self, imagerepo, tag):
        """Pull container with v1 API"""
        Msg().out(f"Info: v1 image id: {imagerepo}", l=Msg.DBG)
        (hdr_data, images_array) = self.get_v1_repo(imagerepo)
        status = self.curl.get_status_code(hdr_data["X-ND-HTTPSTATUS"])
        if status == 401 or not images_array:
            Msg().err("Error: image not found or not authorized")
            return []
        try:
            endpoint = "http://" + hdr_data["x-docker-endpoints"]
        except KeyError:
            endpoint = self.index_url
        (tags_array) = self.get_v1_image_tags(endpoint, imagerepo)
        image_id = self._get_v1_id_from_tags(tags_array, tag)
        if not image_id:
            Msg().err("Error: image tag not found")
            return []
        if len(image_id) <= 8:
            image_id = self._get_v1_id_from_images(images_array, image_id)
            if not image_id:
                Msg().err("Error: image id not found")
                return []
        if not (self.localrepo.setup_tag(tag) and
                self.localrepo.set_version("v1")):
            Msg().err("Error: setting localrepo v1 tag and version")
            return []
        Msg().out("Info: v1 ancestry", image_id, l=Msg.DBG)
        (dummy, ancestry) = self.get_v1_image_ancestry(endpoint, image_id)
        if not ancestry:
            Msg().err("Error: ancestry not found")
            return []
        self.localrepo.save_json("ancestry", ancestry)
        Msg().out("Info: v1 layers", image_id, l=Msg.DBG)
        files = self.get_v1_layers_all(endpoint, ancestry)
        return files

    def _parse_imagerepo(self, imagerepo):
        """Parse imagerepo to extract registry"""
        remoterepo = imagerepo
        registry = ""
        registry_url = ""
        index_url = ""
        components = imagerepo.split('/')
        if '.' in components[0] and len(components) >= 2:
            registry = components[0]
            del components[0]
        elif ('.' not in components[0] and
              components[0] != "library" and len(components) == 1):
            components.insert(0, "library")
        remoterepo = '/'.join(components)
        if registry:
            try:
                registry_url = Config.conf['docker_registries'][registry][0]
                index_url = Config.conf['docker_registries'][registry][1]
            except (KeyError, NameError, TypeError):
                registry_url = registry
                if "://" not in registry:
                    registry_url = f"https://{registry}"
                index_url = registry_url
            if registry_url:
                self.registry_url = registry_url
            if index_url:
                self.index_url = index_url
        return (imagerepo, remoterepo)

    def get(self, imagerepo, tag):
        """Pull a docker image from a v2 registry or v1 index"""
        Msg().out(f"Info: get imagerepo: {imagerepo} tag: {tag}", l=Msg.DBG)
        (imagerepo, remoterepo) = self._parse_imagerepo(imagerepo)
        if self.localrepo.cd_imagerepo(imagerepo, tag):
            new_repo = False
        else:
            self.localrepo.setup_imagerepo(imagerepo)
            new_repo = True
        if self.is_v2():
            files = self.get_v2(remoterepo, tag)  # try v2
        else:
            files = self.get_v1(remoterepo, tag)  # try v1
        if new_repo and not files:
            self.localrepo.del_imagerepo(imagerepo, tag, False)
        return files

    def get_tags(self, imagerepo):
        """List tags from a v2 or v1 repositories"""
        Msg().out("Info: get tags", imagerepo, l=Msg.DBG)
        (dummy, remoterepo) = self._parse_imagerepo(imagerepo)
        if self.is_v2():
            return self.get_v2_image_tags(remoterepo, True)  # try v2
        return self.get_v1_image_tags(remoterepo, True)  # try v1

    def search_init(self, pause):
        """Setup new search"""
        self.search_pause = pause
        self.search_page = 0
        self.search_ended = False

    def search_get_page_v1(self, expression, url):
        """Get search results from Docker Hub using v1 API"""
        if expression:
            url = url + f"/v1/search?q={expression}"
        else:
            url = url + "/v1/search?"
        url += f"&page={str(self.search_page)}"
        (dummy, buf) = self._get_url(url)
        try:
            repo_list = json.loads(buf.getvalue())
            if repo_list["page"] == repo_list["num_pages"]:
                self.search_ended = True
            return repo_list
        except (IOError, OSError, AttributeError,
                ValueError, TypeError):
            self.search_ended = True
            return []

    def search_get_page_v2(self, expression, url, lines=22, official=None):
        """Search results from Docker Hub using v2 API"""
        if not expression:
            expression = '*'
        if expression and official is None:
            url = url + f"/v2/search/repositories?query={expression}"
        elif expression and official is True:
            url = url + f"/v2/search/repositories?query={expression}&is_official=true"
        elif expression and official is False:
            url = url + f"/v2/search/repositories?query={expression}&is_official=false"
        else:
            return []
        url += f"&page_size={str(lines)}"
        if self.search_page != 1:
            url += f"&page={str(self.search_page)}"
        (dummy, buf) = self._get_url(url)
        try:
            repo_list = json.loads(buf.getvalue())
            if repo_list["count"] == self.search_page:
                self.search_ended = True
            return repo_list
        except (IOError, OSError, AttributeError, KeyError,
                ValueError, TypeError):
            self.search_ended = True
            return []

    def search_get_page(self, expression, lines=22):
        """Get search results from Docker Hub"""
        if self.search_ended:
            return []
        self.search_page += 1
        if self.has_search_v2(self.index_url):
            return self.search_get_page_v2(expression, self.index_url, lines)
        if self.has_search_v1():
            return self.search_get_page_v1(expression, self.index_url)
        if self.has_search_v2(self.registry_url):
            return self.search_get_page_v2(expression, self.registry_url, lines)
        if self.has_search_v1(self.registry_url):
            return self.search_get_page_v1(expression, self.registry_url)
        return []


class DockerLocalFileAPI(CommonLocalFileApi):
    """Manipulate Docker container and/or image files"""

    def __init__(self, localrepo):
        CommonLocalFileApi.__init__(self, localrepo)

    def _load_structure(self, tmp_imagedir):
        """Load the structure of a Docker image"""
        structure = {}
        structure["repolayers"] = {}
        structure["repoconfigs"] = {}
        for fname in os.listdir(tmp_imagedir):
            f_path = tmp_imagedir + '/' + fname
            if fname == "repositories":
                structure["repositories"] = \
                        self.localrepo.load_json(f_path)
            elif fname == "manifest.json":
                structure["manifest"] = \
                        self.localrepo.load_json(f_path)
            elif len(fname) >= 69 and fname.endswith(".json"):
                structure["repoconfigs"][fname] = {}
                structure["repoconfigs"][fname]["json"] = \
                        self.localrepo.load_json(f_path)
                structure["repoconfigs"][fname]["json_f"] = \
                        f_path
            elif len(fname) >= 64 and FileUtil(f_path).isdir():
                layer_id = fname
                structure["repolayers"][layer_id] = {}
                for layer_f in os.listdir(f_path):
                    layer_f_path = f_path + '/' + layer_f
                    if layer_f == "VERSION":
                        structure["repolayers"][layer_id]["VERSION"] = \
                                self.localrepo.load_json(layer_f_path)
                    elif layer_f == "json":
                        structure["repolayers"][layer_id]["json"] = \
                                self.localrepo.load_json(layer_f_path)
                        structure["repolayers"][layer_id]["json_f"] = \
                                layer_f_path
                    elif "layer" in layer_f:
                        structure["repolayers"][layer_id]["layer_f"] = \
                                layer_f_path
                    else:
                        Msg().out("Info: warning: unkwnon file in layer:",
                                  f_path, l=Msg.WAR)
        return structure

    def _find_top_layer_id(self, structure, my_layer_id=""):
        """Find the top layer within a Docker image"""
        if "repolayers" not in structure:
            return ""

        if not my_layer_id:
            # if Python 3 TypeError: 'dict_keys' object is not subscriptable
            if sys.version_info[0] >= 3:
                my_layer_id = list(structure["repolayers"].keys())[0]
            else:
                my_layer_id = structure["repolayers"].keys()[0]

        found = ""
        for layer_id in structure["repolayers"]:
            if "parent" not in structure["repolayers"][layer_id]["json"]:
                continue
            if (my_layer_id ==
                    structure["repolayers"][layer_id]["json"]["parent"]):
                found = self._find_top_layer_id(structure, layer_id)
                break

        if not found:
            return my_layer_id

        return found

    def _sorted_layers(self, structure, top_layer_id):
        """Return the layers sorted"""
        sorted_layers = []
        next_layer = top_layer_id
        while next_layer:
            sorted_layers.append(next_layer)
            if "parent" not in structure["repolayers"][next_layer]["json"]:
                break

            next_layer = structure["repolayers"][next_layer]["json"]["parent"]
            if not next_layer:
                break

        return sorted_layers

    def _get_from_manifest(self, structure, imagetag):
        """Search for image:tag in manifest and return Config and layer ids"""
        if "manifest" in structure:
            for repotag in structure["manifest"]:
                if imagetag in repotag["RepoTags"]:
                    layers = []
                    for layer_file in repotag["Layers"]:
                        #layers.append(layer_file.replace("/layer.tar", ""))
                        layers.append(layer_file)
                    layers.reverse()
                    return (repotag["Config"], layers)
        return ("", [])

    def _load_image_step2(self, structure, imagerepo, tag):
        """Load a container image into a repository mimic docker load"""
        imagetag = imagerepo + ':' + tag
        (json_config_file, layers) = \
                self._get_from_manifest(structure, imagetag)
        if json_config_file:
            layer_id = json_config_file.replace(".json", "")
            json_file = structure["repoconfigs"][json_config_file]["json_f"]
            self._move_layer_to_v1repo(json_file, layer_id, "container.json")
        top_layer_id = self._find_top_layer_id(structure)
        layers = self._sorted_layers(structure, top_layer_id)
        for layer_id in layers:
            Msg().out("Info: adding layer:", layer_id, l=Msg.INF)
            if str(structure["repolayers"][layer_id]["VERSION"]) != "1.0":
                Msg().err("Error: layer version unknown")
                return []
            for layer_item in ("json_f", "layer_f"):
                filename = str(structure["repolayers"][layer_id][layer_item])
                if not self._move_layer_to_v1repo(filename, layer_id):
                    Msg().err(f"Error: copying {layer_item[:-2]} file {filename}", l=Msg.VER)
                    return []
        self.localrepo.save_json("ancestry", layers)
        if self._imagerepo:
            imagetag = self._imagerepo + ':' + tag
        return [imagetag]

    def _load_repositories(self, structure):
        """Load other image repositories into this local repo"""
        if "repositories" not in structure:
            return False
        loaded_repositories = []
        for imagerepo in structure["repositories"]:
            for tag in structure["repositories"][imagerepo]:
                if imagerepo and tag:
                    loaded_repo = self._load_image(structure, imagerepo, tag)
                    if loaded_repo:
                        loaded_repositories.extend(loaded_repo)
        return loaded_repositories

    def load(self, tmp_imagedir, imagerepo=None):
        """Load a Docker file created with docker save, mimic Docker
        load. The file is a tarball containing several layers, each
        layer has metadata and data content (directory tree) stored
        as a tar file.
        """
        self._imagerepo = imagerepo
        structure = self._load_structure(tmp_imagedir)
        if not structure:
            Msg().err("Error: failed to load image structure")
            return []
        if "repositories" in structure and structure["repositories"]:
            repositories = self._load_repositories(structure)
        else:
            if not imagerepo:
                imagerepo = Unique().imagename()
            tag = Unique().imagetag()
            repositories = self._load_image(structure, imagerepo, tag)
        return repositories

    def _save_image(self, imagerepo, tag, structure, tmp_imagedir):
        """Prepare structure with metadata for the image"""
        self.localrepo.cd_imagerepo(imagerepo, tag)
        (container_json, layer_files) = self.localrepo.get_image_attributes()
        json_file = tmp_imagedir + "/container.json"
        if (not container_json) or \
           (not self.localrepo.save_json(json_file, container_json)):
            return False
        config_layer_file = str(ChkSUM().sha256(json_file)) + ".json"
        FileUtil(json_file).rename(tmp_imagedir + '/' + config_layer_file)
        manifest_item = {}
        manifest_item["Config"] = config_layer_file
        manifest_item["RepoTags"] = [imagerepo + ':' + tag, ]
        manifest_item["Layers"] = []
        if imagerepo not in structure["repositories"]:
            structure["repositories"][imagerepo] = {}
        parent_layer_id = ""
        for layer_f in layer_files:
            try:
                layer_id = re.search(r"^(?:[^:]+:)?([a-z0-9]+)(?:\.layer)?$",
                                     os.path.basename(layer_f)).group(1)
            except AttributeError:
                return False
            if os.path.exists(tmp_imagedir + "/" + layer_id):
                continue
            FileUtil(tmp_imagedir + "/" + layer_id).mkdir()
            if not FileUtil(layer_f).copyto(tmp_imagedir + "/"
                                            + layer_id + "/layer.tar"):
                return False
            manifest_item["Layers"].append(layer_id + "/layer.tar")
            json_string = self.create_container_meta(layer_id)
            if parent_layer_id:
                json_string["parent"] = parent_layer_id
            else:
                structure["repositories"][imagerepo][tag] = layer_id
            parent_layer_id = layer_id
            if not self.localrepo.save_json(tmp_imagedir + "/"
                                            + layer_id + "/json", json_string):
                return False
            if not FileUtil(tmp_imagedir + "/" +
                            layer_id + "/VERSION").putdata("1.0", 'w'):
                return False
        structure["manifest"].append(manifest_item)
        return True

    def save(self, imagetag_list, imagefile):
        """Save a set of image tags to a file similarly to docker save
        """
        tmp_imagedir = FileUtil("save").mktmp()
        try:
            os.makedirs(tmp_imagedir)
        except (IOError, OSError):
            return False
        structure = {}
        structure["manifest"] = []
        structure["repositories"] = {}
        status = False
        for (imagerepo, tag) in imagetag_list:
            status = self._save_image(imagerepo, tag, structure, tmp_imagedir)
            if not status:
                Msg().err("Error: save image failed:", imagerepo + ':' + tag)
                break
        if status:
            self.localrepo.save_json(tmp_imagedir + "/manifest.json",
                                     structure["manifest"])
            self.localrepo.save_json(tmp_imagedir + "/repositories",
                                     structure["repositories"])
            if not FileUtil(tmp_imagedir).tar(imagefile):
                Msg().err("Error: save image failed in writing tar", imagefile)
                status = False
        else:
            Msg().err("Error: no images specified")
        FileUtil(tmp_imagedir).remove(recursive=True)
        return status



#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
udocker unit tests: CurlHeader
"""

from unittest import TestCase, main
from unittest.mock import patch
from io import StringIO
from udocker.utils.curl import CurlHeader

BUILTINS = "builtins"


class CurlHeaderTestCase(TestCase):
    """Test CurlHeader() http header parser."""

    def test_01_init(self):
        """Test01 CurlHeader() constructor."""
        curl_header = CurlHeader()
        self.assertFalse(curl_header.sizeonly)
        self.assertIsInstance(curl_header.data, dict)
        self.assertEqual("", curl_header.data["X-ND-HTTPSTATUS"])
        self.assertEqual("", curl_header.data["X-ND-CURLSTATUS"])

    def test_02_write(self):
        """Test02 CurlHeader().write()."""
        buff = ["HTTP/1.1 200 OK",
                "Content-Type: application/octet-stream",
                "Content-Length: 32", ]
        curl_header = CurlHeader()
        for line in buff:
            curl_header.write(line)
        status = curl_header.data["content-type"]
        self.assertEqual(status, "application/octet-stream")
        status = curl_header.data["X-ND-HTTPSTATUS"]
        self.assertEqual(status, "HTTP/1.1 200 OK")

        curl_header = CurlHeader()
        for line in buff:
            curl_header.write(line)
        buff_out = curl_header.getvalue()
        self.assertTrue("HTTP/1.1 200 OK" in buff_out)

        line = ""
        curl_header = CurlHeader()
        curl_header.sizeonly = True
        self.assertEqual(-1, curl_header.write(line))

    @patch.object(CurlHeader, 'write')
    def test_03_setvalue_from_file(self, mock_write):
        """Test03 CurlHeader().setvalue_from_file()."""
        fakedata = StringIO('XXXX')
        with patch(BUILTINS + '.open') as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(fakedata.readline, ''))
            curl_header = CurlHeader()
            self.assertTrue(curl_header.setvalue_from_file("filename"))
            self.assertTrue(mock_write.called_with('XXXX'))

    def test_04_getvalue(self):
        """Test04 CurlHeader().getvalue()."""
        curl_header = CurlHeader()
        curl_header.data = "XXXX"
        status = curl_header.getvalue()
        self.assertEqual(status, curl_header.data)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Linux utilities, authentication"""

import re
import pwd
import grp

from udocker.utils.fileutil import FileUtil
from udocker.helper.hostinfo import HostInfo


class NixAuthentication(object):
    """Provides abstraction and useful methods to manage
    passwd and group based authentication, both for the
    host system and for the container.
    If passwd_file and group_file are None then the host
    system authentication databases are used.
    """

    def __init__(self, passwd_file=None, group_file=None,
                 subuid_file=None, subgid_file=None):
        self.passwd_file = passwd_file
        self.group_file = group_file
        self.subuid_file = subuid_file
        self.subgid_file = subgid_file
        if not self.subuid_file:
            self.subuid_file = "/etc/subuid"
        if not self.subgid_file:
            self.subgid_file = "/etc/subgid"

    def _user_in_subid(self, sub_file, wanted_user):
        """get user information from the host /etc/sub*"""
        subid_list = []
        if self.passwd_file:
            (user, dum1, dum2, dum3, dum4, dum5) = \
                    self._get_user_from_file(wanted_user)
        else:
            (user, dum1, dum2, dum3, dum4, dum5) = \
                    self._get_user_from_host(wanted_user)
        try:
            insub = open(sub_file, encoding='utf-8')
        except (IOError, OSError):
            return []
        else:
            for line in insub:
                try:
                    (subuser, subid, count) = line.strip().split(':')
                except ValueError:
                    continue
                if subuser == user:
                    subid_list.extend([(subid, count), ])
            insub.close()
        return subid_list

    def user_in_subuid(self, wanted_user):
        """Find if wanted_user is in /etc/subuid"""
        return self._user_in_subid(self.subuid_file, wanted_user)

    def user_in_subgid(self, wanted_user):
        """Find if wanted_user is in /etc/subgid"""
        return self._user_in_subid(self.subgid_file, wanted_user)

    def _get_user_from_host(self, wanted_user):
        """get user information from the host /etc/passwd"""
        wanted_uid = ""
        if (isinstance(wanted_user, int) or
                re.match("^\\d+$", wanted_user)):
            wanted_uid = str(wanted_user)
            wanted_user = ""

        if wanted_uid:
            try:
                usr = pwd.getpwuid(int(wanted_uid))
            except (IOError, OSError, KeyError):
                return ("", "", "", "", "", "")
            return (str(usr.pw_name), str(usr.pw_uid), str(usr.pw_gid),
                    str(usr.pw_gecos), usr.pw_dir, usr.pw_shell)
        try:
            usr = pwd.getpwnam(wanted_user)
        except (IOError, OSError, KeyError):
            return ("", "", "", "", "", "")
        return (str(usr.pw_name), str(usr.pw_uid), str(usr.pw_gid),
                str(usr.pw_gecos), usr.pw_dir, usr.pw_shell)

    def _get_group_from_host(self, wanted_group):
        """get group information from the host /etc/group"""
        wanted_gid = ""
        if (isinstance(wanted_group, int) or
                re.match("^\\d+$", wanted_group)):
            wanted_gid = str(wanted_group)
            wanted_group = ""
        if wanted_gid:
            try:
                hgr = grp.getgrgid(int(wanted_gid))
            except (IOError, OSError, KeyError):
                return ("", "", "")
            return (str(hgr.gr_name), str(hgr.gr_gid), str(hgr.gr_mem))

        try:
            hgr = grp.getgrnam(wanted_group)
        except (IOError, OSError, KeyError):
            return ("", "", "")
        return (str(hgr.gr_name), str(hgr.gr_gid), str(hgr.gr_mem))

    def _get_user_from_file(self, wanted_user):
        """Get user from a passwd file"""
        wanted_uid = ""
        if (isinstance(wanted_user, int) or
                re.match("^\\d+$", wanted_user)):
            wanted_uid = str(wanted_user)
            wanted_user = ""
        try:
            inpasswd = open(self.passwd_file, encoding='utf-8')
        except (IOError, OSError):
            return ("", "", "", "", "", "")
        else:
            for line in inpasswd:
                (user, dummy, uid, gid, gecos, home,
                 shell) = line.strip().split(':')
                if wanted_user and user == wanted_user:
                    return (user, uid, gid, gecos, home, shell)
                if wanted_uid and uid == wanted_uid:
                    return (user, uid, gid, gecos, home, shell)
            inpasswd.close()
            return ("", "", "", "", "", "")

    def _get_group_from_file(self, wanted_group):
        """Get group from a group file"""
        wanted_gid = ""
        if (isinstance(wanted_group, int) or
                re.match("^\\d+$", wanted_group)):
            wanted_gid = str(wanted_group)
            wanted_group = ""
        try:
            ingroup = open(self.group_file, encoding='utf-8')
        except (IOError, OSError):
            return ("", "", "")
        else:
            for line in ingroup:
                (group, dummy, gid, users) = line.strip().split(':')
                if wanted_group and group == wanted_group:
                    return (group, gid, users)
                if wanted_gid and gid == wanted_gid:
                    return (group, gid, users)
            ingroup.close()
            return ("", "", "")

    def add_user(self, user, passw, uid, gid, gecos,
                 home, shell):
        """Add a *nix user to a /etc/passwd file"""
        line = f"{user}:{passw}:{uid}:{gid}:{gecos}:{home}:{shell}\n"
        if line in FileUtil(self.passwd_file).getdata('r'):
            return True
        return FileUtil(self.passwd_file).putdata(line, 'a')

    def add_group(self, group, gid, users=None):
        """Add a group to a /etc/passwd file"""
        users_str = ""
        if isinstance(users, list):
            for username in users:
                users_str += f"{username},"
        line = f"{group}:x:{gid}:{users_str}\n"
        if line in FileUtil(self.group_file).getdata('r'):
            return True
        return FileUtil(self.group_file).putdata(line, 'a')

    def get_user(self, wanted_user):
        """Get host or container user"""
        if self.passwd_file:
            return self._get_user_from_file(wanted_user)
        return self._get_user_from_host(wanted_user)

    def get_group(self, wanted_group):
        """Get host or container group"""
        if self.group_file:
            return self._get_group_from_file(wanted_group)
        return self._get_group_from_host(wanted_group)

    def get_home(self):
        """Get host or container home directory"""
        (r_user, dum1, dum2, dum3, r_home, dum4) = self.get_user(HostInfo.uid)
        if r_user:
            return r_home
        return ""
